import abc
from plico.utils.decorator import returnsNone, returns, returnsForExample
from palpao.types.deformable_mirror_status import DeformableMirrorStatus
from six import with_metaclass




class SnapshotEntry(object):
    COMMAND_COUNTER= "COMMAND_COUNTER"
    SERIAL_NUMBER= "SERIAL_NUMBER"
    STEP_COUNTER= "STEP_COUNTER"



class AbstractDeformableMirrorClient(with_metaclass(abc.ABCMeta, object)):
    """
    Interface to control a deformable mirror


    Assume a modal control of the DM.

    """

    @abc.abstractmethod
    def getNumberOfModes(self):
        """ Number of modes of the deformable mirror

        Return the number of modes of the deformable mirror.
        number of degrees of freedom.

        Return:
            numberOfModes (int): the number of modes of the deformable mirror.
        """
        assert False


    @abc.abstractmethod
    @returnsNone
    def setShape(self, command):
        """ Set Deformable Mirror Shape

        Send to the controller the request to set the DM shape

        Parameters:
            command (:obj:ndarray): an array containing the required value for the actuators/modes
                The size of the array must be equal to the number of modes of the DM


        """
        assert False


    @abc.abstractmethod
    @returnsNone
    def getShape(self):
        """ Get Deformable Mirror Shape

        Send to the controller the request to set the DM shape

        Return:
            shape (:obj:ndarray): an array containing the measured value for the actuators/modes
                The size of the array must be equal to the number of modes of the DM

        The value are measured if the DM has an internal metrology on position
        The shape sequence is taken into account.

        """
        assert False


    @abc.abstractmethod
    def loadShapeSequence(self, shapeSequence, timeStepInSeconds):
        """ Load a shape sequence to be applied to the mirror

        Every element of the sequence correspond to a DM shape
        After timeStepInSeconds the next element of the sequence is applied
        The sequence is executed ciclycally until 

        The shape sequence is added to the current shape
        The shape sequence is applied as a circular buffer 

        Parameters:
            shapeSequence (:obj:ndarray): an array containing the shape sequence value for the actuators/modes
                The array size is (nModes, nTimeSteps)
            timeStepInSeconds (float): updating interval of sequence.
                Every timeStepInSeconds the controller applies the next column of shapeSequence
        """
        assert False


    @abc.abstractmethod
    @returnsNone
    def startShapeSequence(self):
        assert False


    @abc.abstractmethod
    @returnsNone
    def stopShapeSequence(self):
        assert False


    @abc.abstractmethod
    @returnsNone
    def enableControlLoop(self, boolEnableOrDisable):
        """ Enable control loop

        If the deformable mirror controller has position feedback, 
        enable the position control loop
        Else, raise TypeError
        """
        assert False


    @abc.abstractmethod
    @returnsForExample({'WFS_CAMERA.EXPOSURE_TIME_MS': 10})
    def getSnapshot(self, prefix):
        assert False


    @abc.abstractmethod
    @returns(DeformableMirrorStatus)
    def getStatus(self):
        assert False



#!/usr/bin/env python
import unittest
import numpy as np
from palpao.types.deformable_mirror_status import DeformableMirrorStatus



class DeformableMirrorStatusTest(unittest.TestCase):


    def testHappyPath(self):
        numberOfActs= 10
        numberOfModes= 8
        actuatorCommands= np.arange(numberOfActs)
        commandCounter= 42
        status= DeformableMirrorStatus(
            numberOfActs,
            numberOfModes,
            actuatorCommands,
            commandCounter)

        self.assertEqual(numberOfActs, status.numberOfActuators())
        self.assertEqual(numberOfModes, status.numberOfModes())
        self.assertTrue(np.allclose(
            actuatorCommands, status.actuatorCommands()))
        self.assertEqual(commandCounter, status.commandCounter())


if __name__ == "__main__":
    unittest.main()



from plico.utils.decorator import override, returnsNone, returns
from palpao.types.modal_basis import ModalBasis
from palpao.calibration.abstract_calibration_manager import \
    AbstractCalibrationManager



__version__= "$Id: in_memory_calibration_manager.py 27 2018-01-27 08:48:07Z lbusoni $"


class InMemoryCalibrationManager(AbstractCalibrationManager):

    def __init__(self):
        self._modalBasisDict= {}


    @override
    @returnsNone
    def saveModalBasis(self, tag, modalBasis):
        self._modalBasisDict[tag]= modalBasis


    @override
    @returns(ModalBasis)
    def loadModalBasis(self, tag):
        return self._modalBasisDict[str(tag)]



import numpy as np
from plico.utils.decorator import override, returns
from palpao.client.abstract_deformable_mirror_client import \
    AbstractDeformableMirrorClient
from palpao.utils.timeout import Timeout
import time
from palpao.types.deformable_mirror_status import DeformableMirrorStatus



class SimulatedDeformableMirrorClient(AbstractDeformableMirrorClient):

    N_MODES= 4

    def __init__(self, timeModule=time):
        self._timeMod= timeModule
        self._shape= np.zeros(self.N_MODES)
        self._commandCounter= 0
        self._isControlLoopEnabled= False
        self._isShapeSequenceEnabled= False
        self._shapeSeq= np.zeros(self.N_MODES).reshape((
            self.N_MODES, 1))
        self._nElementsShapeSeq= 1
        self._seqTimeStepInSeconds= 1
        self._shapeSeqIdx= 0
        self._timeStampSequence= 0


    @override
    def enableControlLoop(self,
                          boolEnableOrDisable,
                          timeoutInSec=Timeout.GENERIC_COMMAND):
        self._isControlLoopEnabled= boolEnableOrDisable


    @override
    def loadShapeSequence(self,
                          shapeSequence,
                          timeStepInSeconds,
                          timeoutInSec=Timeout.GENERIC_COMMAND):
        self._shapeSeq= shapeSequence
        self._seqTimeStepInSeconds= timeStepInSeconds
        self._shapeSeqIdx= 0
        self._nElementsShapeSeq= self._shapeSeq.shape[1]


    @override
    def startShapeSequence(self,
                           timeoutInSec=Timeout.GENERIC_COMMAND):
        self._isShapeSequenceEnabled= True
        self._timeStampSequence= self._timeMod.time()


    @override
    def stopShapeSequence(self,
                          timeoutInSec=Timeout.GENERIC_COMMAND):
        self._isShapeSequenceEnabled= False


    @override
    def setShape(self,
                 command,
                 timeoutInSec=Timeout.MIRROR_SET_SHAPE):
        self._shape= command.copy()
        self._commandCounter+= 1


    @override
    def getShape(self, timeoutInSec=Timeout.MIRROR_GET_SHAPE):
        shape= self._shape.copy()
        if self._isShapeSequenceEnabled:
            now= self._timeMod.time()
            nSteps= int((now - self._timeStampSequence) /
                        self._seqTimeStepInSeconds)
            seqIdx= nSteps % self._nElementsShapeSeq
            shape += self._shapeSeq[:, seqIdx]
        return shape


    @override
    def getNumberOfModes(self, timeoutInSec=Timeout.GENERIC_COMMAND):
        return self.N_MODES


    @override
    @returns(DeformableMirrorStatus)
    def getStatus(self, timeoutInSec=Timeout.MIRROR_GET_STATUS):
        return DeformableMirrorStatus(self._shape,
                                      self._commandCounter)



    @override
    def getSnapshot(self):
        return {}



#!/usr/bin/env python

from plico.client.hackerable_client import HackerableClient
from palpao.client.abstract_deformable_mirror_client import \
    AbstractDeformableMirrorClient
from plico.rpc.abstract_remote_procedure_call import \
    AbstractRemoteProcedureCall
from plico.utils.logger import Logger
from plico.utils.decorator import override, returns
from palpao.utils.timeout import Timeout
from palpao.types.deformable_mirror_status import DeformableMirrorStatus
from plico.client.serverinfo_client import ServerInfoClient




class DeformableMirrorClient(AbstractDeformableMirrorClient,
                             HackerableClient, ServerInfoClient):

    def __init__(self,
                 rpcHandler,
                 sockets):
        assert isinstance(rpcHandler, AbstractRemoteProcedureCall)

        self._rpcHandler= rpcHandler
        self._requestSocket= sockets.serverRequest()
        self._statusSocket= sockets.serverStatus()
        self._logger= Logger.of('DeformableMirrorClient')

        HackerableClient.__init__(self,
                                  self._rpcHandler,
                                  self._requestSocket,
                                  self._logger)
        ServerInfoClient.__init__(self,
                                  self._rpcHandler,
                                  self._requestSocket,
                                  self._logger)


    @override
    def enableControlLoop(self,
                          boolEnableOrDisable,
                          timeoutInSec=Timeout.GENERIC_COMMAND):
        return self._rpcHandler.sendRequest(
            self._requestSocket,
            'enableControlLoop', [boolEnableOrDisable],
            timeout=timeoutInSec)



    @override
    def loadShapeSequence(self,
                          shapeSequence,
                          timeStepInSeconds,
                          timeoutInSec=Timeout.GENERIC_COMMAND):
        return self._rpcHandler.sendRequest(
            self._requestSocket,
            'loadShapeSequence', [shapeSequence, timeStepInSeconds],
            timeout=timeoutInSec)


    @override
    def startShapeSequence(self,
                           timeoutInSec=Timeout.GENERIC_COMMAND):
        return self._rpcHandler.sendRequest(
            self._requestSocket,
            'startShapeSequence', [],
            timeout=timeoutInSec)


    @override
    def stopShapeSequence(self,
                          timeoutInSec=Timeout.GENERIC_COMMAND):
        return self._rpcHandler.sendRequest(
            self._requestSocket,
            'stopShapeSequence', [],
            timeout=timeoutInSec)


    @override
    def setShape(self,
                 command,
                 timeoutInSec=Timeout.MIRROR_SET_SHAPE):
        return self._rpcHandler.sendRequest(
            self._requestSocket,
            'setShape', [command],
            timeout=timeoutInSec)


    @override
    def getShape(self, timeoutInSec=Timeout.MIRROR_GET_SHAPE):
        return self._rpcHandler.sendRequest(
            self._requestSocket,
            'getShape', [],
            timeout=timeoutInSec)


    @override
    def getNumberOfModes(self, timeoutInSec=Timeout.GENERIC_COMMAND):
        return int(self.getStatus(timeoutInSec).numberOfModes())


    @override
    @returns(DeformableMirrorStatus)
    def getStatus(self, timeoutInSec=Timeout.MIRROR_GET_STATUS):
        return self._rpcHandler.receivePickable(
            self._statusSocket,
            timeoutInSec)


    @override
    def getSnapshot(self,
                    prefix,
                    timeoutInSec=Timeout.MIRROR_GET_STATUS):
        return self._rpcHandler.sendRequest(
            self._requestSocket,
            'getSnapshot', [prefix],
            timeout=timeoutInSec)




__version__ = "$Id: constants.py 45 2018-04-21 17:56:18Z lbusoni $"


class Constants:
    METER_2_NANOMETER= 1e9
    APP_NAME= "inaf.arcetri.ao.palpao"
    APP_AUTHOR= "INAF Arcetri Adaptive Optics"
    THIS_PACKAGE= 'palpao'





import os
import pyfits
from palpao.calibration.abstract_calibration_manager import \
    AbstractCalibrationManager
from plico.utils.decorator import override, returnsNone, returns, cacheResult
from palpao.types.modal_basis import ModalBasis
from plico.utils.fits_file_based_calibration_manager \
    import FitsFileBasedCalibrationManager



class CalibrationManagerException(Exception):
    pass


class CalibrationManager(AbstractCalibrationManager,
                         FitsFileBasedCalibrationManager):

    def __init__(self, calibrationRootDir):
        self._calibRootDir = calibrationRootDir


    def _checkTag(self, tag):
        if tag is None:
            raise CalibrationManagerException(
                "A tag must be given but is None")
        if len(tag) == 0:
            raise CalibrationManagerException(
                "A tag name must be valid but it is '%s'" % (tag))


    def getModalBasisFileName(self, tag):
        return os.path.join(self._calibRootDir,
                            "modal_basis",
                            "%s.fits" % tag)


    @override
    @returnsNone
    def saveModalBasis(self, tag, modalBasis):
        self._checkTag(tag)
        fileName= self.getModalBasisFileName(tag)
        self._createFoldersIfMissing(fileName)
        pyfits.writeto(fileName,
                       modalBasis.modalToZonalMatrix,
                       clobber=False)


    @override
    @returns(ModalBasis)
    @cacheResult
    def loadModalBasis(self, tag):
        self._checkTag(tag)
        fileName= self.getModalBasisFileName(tag)
        hduList= pyfits.open(fileName)
        return ModalBasis(hduList[0].data)



    def getPiTipTiltCalibrationFileName(self):
        return os.path.join(self._calibRootDir,
                            "modulator/pi_calibration.py")


    def _loadPiTipTiltCalibrationModule(self):
        import imp
        mm= imp.load_source('pi_calibration',
                            self.getPiTipTiltCalibrationFileName())
        return mm.PhysikInstrumenteCalibration()


    @override
    @cacheResult
    def loadPiTipTiltCalibration(self, serialNumber):
        pic= self._loadPiTipTiltCalibrationModule()
        return pic.getCalibrationFor(serialNumber)



import abc
from plico.utils.decorator import returns, returnsNone
from palpao.types.modal_basis import ModalBasis
from six import with_metaclass


__version__= "$Id: abstract_calibration_manager.py 27 2018-01-27 08:48:07Z lbusoni $"



class AbstractCalibrationManager(with_metaclass(abc.ABCMeta, object)):

    @abc.abstractmethod
    @returns(ModalBasis)
    def loadModalBasis(self, tag):
        assert False


    @abc.abstractmethod
    @returnsNone
    def saveModalBasis(self, tag, modalBasis):
        assert False





#!/usr/bin/env python
import unittest
import numpy as np
from plico.utils.decorator import override
from plico.rpc.dummy_remote_procedure_call import DummyRpcHandler
from plico.rpc.dummy_sockets import DummySockets
from palpao.client.deformable_mirror_client import DeformableMirrorClient
from palpao.utils.timeout import Timeout
from palpao.types.deformable_mirror_status import DeformableMirrorStatus



class TesterRpcHandler(DummyRpcHandler):

    def __init__(self):
        self._sendRequestHistory= []
        self._receivePickableHistory= []
        self._sendRequestCounter= 0
        self._receivePickableCounter= 0


    @override
    def sendRequest(self, socket, command, args, timeout=1):
        self._sendRequestHistory.append(
            (socket, command, args, timeout))
        self._sendRequestCounter += 1


    @override
    def receivePickable(self, socket, timeout=1):
        self._receivePickableCounter+= 1
        self._receivePickableHistory.append(
            (socket, timeout))
        return self._objToReturnWithReceivePickable


    def getLastSendRequestArguments(self):
        return self._sendRequestHistory[-1]


    def getLastReceivePickableArguments(self):
        return self._receivePickableHistory[-1]


    def wantsPickable(self, objToReturnWithReceivePickable):
        self._objToReturnWithReceivePickable= \
            objToReturnWithReceivePickable


class MyRpcHandler(TesterRpcHandler):
    pass


class MySockets(DummySockets):
    pass


class DeformableMirrorClientTest(unittest.TestCase):


    def setUp(self):
        self._rpc= MyRpcHandler()
        self._sockets= MySockets()
        self._client= DeformableMirrorClient(self._rpc, self._sockets)


    def tearDown(self):
        pass


    def testSetShape(self):
        mirrorWantedShape= np.identity(4)
        self._client.setShape(mirrorWantedShape)
        self.assertEqual(
            self._rpc.getLastSendRequestArguments(),
            (self._sockets.serverRequest(),
             'setShape',
             [mirrorWantedShape],
             Timeout.MIRROR_SET_SHAPE))


    def testGetPosition(self):
        timeoutInSec= 12
        _= self._client.getShape(timeoutInSec)
        self.assertEqual(
            self._rpc.getLastSendRequestArguments(),
            (self._sockets.serverRequest(),
             'getShape',
             [],
             timeoutInSec))


    def testGetStatus(self):
        wantedInstrumentStatus= DeformableMirrorStatus(
            12, 34, 'foo', 'tux')
        self._rpc.wantsPickable(wantedInstrumentStatus)

        timeoutInSec= 22
        gotInstrumentStatus= self._client.getStatus(timeoutInSec)
        self.assertEqual(
            wantedInstrumentStatus, gotInstrumentStatus)

        self.assertEqual(
            self._rpc.getLastReceivePickableArguments(),
            (self._sockets.serverStatus(),
             timeoutInSec))


if __name__ == "__main__":
    unittest.main()






class DeformableMirrorStatus(object):

    def __init__(self,
                 numberOfActuators,
                 numberOfModes,
                 actuatorCommands,
                 commandCounter):
        self._numberOfActuators= numberOfActuators
        self._numberOfModes= numberOfModes
        self._actuatorCommands= actuatorCommands
        self._commandCounter= commandCounter


    def commandCounter(self):
        return self._commandCounter


    def actuatorCommands(self):
        return self._actuatorCommands


    def numberOfActuators(self):
        return self._numberOfActuators


    def numberOfModes(self):
        return self._numberOfModes


    def __repr__(self):
        stra= "Number of Actuators: %s\n" % self.numberOfActuators()
        stra+= "Number of Modes: %s\n" % self.numberOfModes()
        stra+= "Actuator commands: %s\n" % self.actuatorCommands()
        stra+= "command counter: %s\n" % self.commandCounter()
        return stra



#!/usr/bin/env python
import os
import unittest
import shutil

import numpy as np
from palpao.calibration.calibration_manager import CalibrationManager,\
    CalibrationManagerException
from palpao.types.modal_basis import ModalBasis


__version__= "$Id: calibration_manager_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class CalibrationManagerTest(unittest.TestCase):

    CALIB_DIR= "./calib_tmp"


    def _removeCalibrationDir(self):
        if os.path.exists(self.CALIB_DIR):
            shutil.rmtree(self.CALIB_DIR)


    def setUp(self):
        self._removeCalibrationDir()
        self.calibMgr= CalibrationManager(self.CALIB_DIR)


    def tearDown(self):
        self._removeCalibrationDir()


    def _createModalBasis(self):
        return ModalBasis(np.arange(6).reshape((3, 2)))


    def testStorageOfModalBasis(self):
        result= self._createModalBasis()

        self.calibMgr.saveModalBasis("foo", result)
        self.assertTrue(os.path.exists(
            os.path.join(self.CALIB_DIR, "modal_basis", "foo.fits")))

        loaded= self.calibMgr.loadModalBasis("foo")
        self.assertTrue(np.array_equal(
            result.modalToZonalMatrix, loaded.modalToZonalMatrix))


    def testInvalidTag(self):
        res= self._createModalBasis()
        self.assertRaises(
            CalibrationManagerException,
            self.calibMgr.saveModalBasis,
            None, res)
        self.assertRaises(
            CalibrationManagerException,
            self.calibMgr.saveModalBasis,
            "", res)

        self.assertRaises(
            CalibrationManagerException,
            self.calibMgr.loadModalBasis, None)
        self.assertRaises(
            CalibrationManagerException,
            self.calibMgr.loadModalBasis, "")





if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
import unittest
import numpy as np
from palpao.client.simulated_deformable_mirror_client import \
    SimulatedDeformableMirrorClient
from test.fake_time_mod import FakeTimeMod


class SimulatedDeformableMirrorClientTest(unittest.TestCase):


    def setUp(self):
        self.timeMod= FakeTimeMod()
        self.dm= SimulatedDeformableMirrorClient(timeModule=self.timeMod)
        self.nModes= SimulatedDeformableMirrorClient.N_MODES


    def testGetNumberOfModes(self):
        nModes= self.dm.getNumberOfModes()
        self.assertEqual(
            SimulatedDeformableMirrorClient.N_MODES, nModes)


    def testSetAndGetShape(self):
        wantShape=np.arange(self.nModes)
        self.dm.setShape(wantShape)
        getShape= self.dm.getShape()
        self.assertTrue(np.array_equal(wantShape, getShape))


    def testLoopSequence(self):
        timeStepInSeconds=0.1
        seqNumberOfTimeSteps= 100

        initialShape= np.arange(self.nModes) * 42
        self.dm.setShape(initialShape)
        seq= np.arange(self.nModes * seqNumberOfTimeSteps).reshape(
            self.nModes, seqNumberOfTimeSteps)
        self.dm.loadShapeSequence(seq, timeStepInSeconds)

        self.dm.startShapeSequence()
        shapeBefore= self.dm.getShape()
        self.timeMod.sleep(timeStepInSeconds * 2)
        shapeAfter= self.dm.getShape()
        self.assertFalse(np.array_equal(shapeAfter, shapeBefore))

        self.dm.stopShapeSequence()
        shapeBefore= self.dm.getShape()
        self.timeMod.sleep(timeStepInSeconds * 2)
        shapeAfter= self.dm.getShape()
        self.assertTrue(np.array_equal(shapeAfter, shapeBefore))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest
import numpy as np
from palpao.calibration.in_memory_calibration_manager import \
    InMemoryCalibrationManager
from palpao.types.modal_basis import ModalBasis

__version__= "$Id: in_memory_calibration_manager_test.py 26 2018-01-26 19:06:25Z lbusoni $"


class InMemoryCalibrationManagerTest(unittest.TestCase):

    def setUp(self):
        self.calibMgr= InMemoryCalibrationManager()


    def testModalBasisStorage(self):
        tag= "20140909_110800"
        originalObject= ModalBasis(np.arange(8).reshape((4, 2)))
        self.calibMgr.saveModalBasis(tag, originalObject)
        storedObject= self.calibMgr.loadModalBasis(tag)
        self.assertEqual(storedObject, originalObject)



if __name__ == "__main__":
    unittest.main()
